import dash
import dash_html_components as html
from nwb_web_gui.dashapps.pages.viewer.viewer import Viewer
import dash_bootstrap_components as dbc
from nwb_web_gui.dashapps.components.navbar import render_navbar
from dash.dependencies import Input, Output, State, ALL, MATCH
from flask.helpers import get_root_path
from pathlib import Path

def init_viewer(server):
    FONT_AWESOME = "https://use.fontawesome.com/releases/v5.7.2/css/all.css"
    external_stylesheets = [dbc.themes.BOOTSTRAP, FONT_AWESOME]
    assets = Path(get_root_path(__name__)).parent.parent.parent / 'static'

    dash_app = dash.Dash(
        __name__,
        server=server,
        external_stylesheets=external_stylesheets,
        suppress_callback_exceptions=True,
        routes_pathname_prefix='/viewer/',
        assets_folder=assets
    )

    navbar = render_navbar()
    viewer = Viewer(parent_app=dash_app)
    # Create Dash Layout
    dash_app.layout = html.Div([navbar, viewer])

    return dash_app.server


import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output
from flask import current_app as app


'''
Collor pallete of icon:
orange: #d17128
grey: #8f8f8f
blue: #114b7c
'''

NAV_LOGO = "assets/logo_nwb.png"


def render_navbar():
    """Make Navbar"""
    apps_dict = {
        'converter': {
            'render': app.config['NWB_GUI_RENDER_CONVERTER'],
            'id': 'nav_nwb_converter',
            'children': 'NWB Converter',
            'href': '/converter'
        },
        'viewer': {
            'render': app.config['NWB_GUI_RENDER_VIEWER'],
            'id': 'nav_nwb_viewer',
            'children': 'NWB Viewer',
            'href': '/viewer'
        },
        'dashboard': {
            'render': app.config['NWB_GUI_RENDER_DASHBOARD'],
            'id': 'nav_nwb_dashboard',
            'children': 'Dashboard',
            'href': '/dashboard'
        },
        "close_gui": {
            "render": True,
            'id': "close_gui",
            "children": "Close GUI",
            "href": '/shutdown'
        }
    }

    navchildren = []
    for k, v in apps_dict.items():
        if v['render']:
            item = dbc.NavItem(html.A(
                            id=v['id'], children=v['children'], href=v['href'],
                            style={"font-size": "120%", "font-weight": "normal"}, className='nav-link'
                    ))
            navchildren.append(item)

    navbar = dbc.Navbar(
        [
            html.A(
                dbc.Row(
                    [
                        dbc.Col(html.Img(src=NAV_LOGO, height="60px")),
                        # dbc.Col(dbc.NavbarBrand("Navbar", className="ml-2")),
                    ],
                    align="center",
                    no_gutters=True,
                ),
                href="/",
                id="nav_brand"
            ),
            dbc.Nav(
                [
                    dbc.Container(
                        navchildren
                    )
                ],
                horizontal='end',
                className="ml-auto flex-nowrap mt-3 mt-md-0"
            )
        ],
        color="dark",
        dark=True
    )

    return navbar



from flask import Flask
import os
from distutils.util import strtobool
import importlib
from .config import config


def init_app():
    """Construct core Flask application with embedded Dash app."""
    app = Flask(__name__, instance_relative_config=False, template_folder='templates')

    app.config.from_object(config.get(os.environ.get('FLASK_CONFIG') or 'default'))

    # Variables from ENV vars
    app.config['NWB_GUI_CONVERTER_MODULE'] = os.environ.get('NWB_GUI_CONVERTER_MODULE')
    app.config['NWB_GUI_CONVERTER_CLASS'] = os.environ.get('NWB_GUI_CONVERTER_CLASS')
    app.config['NWB_GUI_DASHBOARD_MODULE'] = os.environ.get('NWB_GUI_DASHBOARD_MODULE')
    app.config['NWB_GUI_DASHBOARD_CLASS'] = os.environ.get('NWB_GUI_DASHBOARD_CLASS')
    app.config['NWB_GUI_ROOT_PATH'] = os.environ.get('NWB_GUI_ROOT_PATH')
    app.config['NWB_GUI_SECRET_KEY'] = os.environ.get('NWB_GUI_SECRET_KEY')
    app.config['NWB_GUI_RENDER_CONVERTER'] = bool(strtobool(os.environ.get('NWB_GUI_RENDER_CONVERTER', 'True')))
    app.config['NWB_GUI_RENDER_VIEWER'] = bool(strtobool(os.environ.get('NWB_GUI_RENDER_VIEWER', 'False')))
    app.config['NWB_GUI_RENDER_DASHBOARD'] = bool(strtobool(os.environ.get('NWB_GUI_RENDER_DASHBOARD', 'False')))

    with app.app_context():

        # Import NWB converter
        if app.config['NWB_GUI_CONVERTER_CLASS'] == 'example':
            from .dashapps.pages.converter.converter_utils.converter_example import ExampleNWBConverter
            converter_class = ExampleNWBConverter
        else:
            converter_class = getattr(
                importlib.import_module(app.config['NWB_GUI_CONVERTER_MODULE']),
                app.config['NWB_GUI_CONVERTER_CLASS']
            )

        # Import Dash application
        if app.config['NWB_GUI_RENDER_CONVERTER']:
            from .dashapps.pages.converter.init_coverter import init_converter
            global converter
            converter = init_converter(server=app, converter_class=converter_class)
        if app.config['NWB_GUI_RENDER_VIEWER']:
            from .dashapps.pages.viewer.init_viewer import init_viewer
            global viewer
            init_viewer(app)
        if app.config['NWB_GUI_RENDER_DASHBOARD']:
            from .dashapps.pages.dashboard.init_dashboard import init_dashboard
            dashboard_class = getattr(
                importlib.import_module(app.config['NWB_GUI_DASHBOARD_MODULE']),
                app.config['NWB_GUI_DASHBOARD_CLASS']
            )
            global dashboard
            dashboard = init_dashboard(app, dashboard_class=dashboard_class)

        # Import parts of our core Flask app
        from . import routes

        return app



from . import init_app
from pathlib import Path
from threading import Timer
import webbrowser
import os


def parse_arguments():
    """
    Command line shortcut to open GUI editor.
    Usage:
    $ nwb-web-gui [--converter] [--data_path] [--port] [--dev] [--render_dashboard] [--render_viewer] [--render_converter]

    converter : str
        Optional.
        The name of the specfic converter module to be used.
        If not specified will be used example converter
    data_path : str
        Optional. Base path to experimental data.
    port : int
        Optional. Port where app will be running.
    render_dashboard : bool
        Optional. Render dashboard page
    render_viewer : bool
        Optional. Render viewer page
    render_converter : bool
        Optional. Render converter page
    """
    import argparse

    parser = argparse.ArgumentParser(
        description='NWB WEB GUI,',
    )

    parser.add_argument(
        "--converter",
        default='example',
        help="Converter module name to be used"
    )
    parser.add_argument(
        "--data_path",
        default='.',
        help="Path to datasets. Defaults to current working directory."
    )
    parser.add_argument(
        "--port",
        default='5000',
        help="Port where app will be running. Defaults to 5000."
    )
    parser.add_argument(
        "--dev",
        default=False,
        help="Run in development mode. Defaults to False."
    )
    parser.add_argument(
        "--render_dashboard",
        default='False',
        help="Render dashboard page"
    )
    parser.add_argument(
        "--render_viewer",
        default='False',
        help="Render viewer page"
    )
    parser.add_argument(
        "--render_converter",
        default='True',
        help="Render coverter page"
    )

    # Parse arguments
    args = parser.parse_args()

    return args


def cmd_line_shortcut():
    run_args = parse_arguments()

    # Set ENV variables for app
    data_path = str(Path(run_args.data_path))
    render_converter = run_args.render_converter
    render_viewer = run_args.render_viewer
    render_dashboard = run_args.render_dashboard
    converter = run_args.converter

    os.environ['FLASK_ENV'] = 'production'
    os.environ['NWB_GUI_ROOT_PATH'] = data_path
    os.environ['NWB_GUI_RENDER_CONVERTER'] = render_converter
    os.environ['NWB_GUI_RENDER_VIEWER'] = render_viewer
    os.environ['NWB_GUI_RENDER_DASHBOARD'] = render_dashboard

    # Choose converter
    os.environ['NWB_GUI_CONVERTER_MODULE'] = converter
    if converter == 'example':
        os.environ['NWB_GUI_CONVERTER_CLASS'] = converter

    print(f'NWB GUI running on localhost:{run_args.port}')
    print(f'Data path: {data_path}')
    if run_args.dev:
        os.environ['FLASK_ENV'] = 'development'
        print('Running in development mode')

    # Initialize app
    app = init_app()

    # Open browser after 1 sec
    def open_browser():
        webbrowser.open_new(f'http://localhost:{run_args.port}/')

    Timer(1, open_browser).start()

    # Run app
    app.run(
        host='0.0.0.0',
        port=run_args.port,
        debug=run_args.dev,
        use_reloader=run_args.dev
    )



import dash
import dash_html_components as html
from nwb_web_gui.dashapps.pages.converter.converter import ConverterForms
import dash_bootstrap_components as dbc
from nwb_web_gui.dashapps.components.navbar import render_navbar
from flask.helpers import get_root_path
from pathlib import Path


def init_converter(server, converter_class):
    FONT_AWESOME = "https://use.fontawesome.com/releases/v5.7.2/css/all.css"
    external_stylesheets = [dbc.themes.BOOTSTRAP, FONT_AWESOME]
    #assets = Path(get_root_path(__name__)).parent.parent.parent.parent / 'static'
    assets = Path(get_root_path(__name__)).parent.parent.parent / 'static'

    dash_app = dash.Dash(
        __name__,
        server=server,
        external_stylesheets=external_stylesheets,
        suppress_callback_exceptions=True,
        routes_pathname_prefix='/converter/',
        assets_folder=assets
    )

    navbar = render_navbar()
    # Create Dash Layout
    dash_app.layout = html.Div([
        navbar,
        ConverterForms(parent_app=dash_app, converter_class=converter_class)
    ])
    # dash_app.enable_dev_tools(debug=True)
    dash_app.enable_dev_tools(debug=False)

    return dash_app.server



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

with open('requirements.txt', 'r') as f:
    install_requires = [w for w in f.read().strip().split('\n') if not w.startswith('-e')]

setup(
    name='nwb-web-gui',
    version='0.1.3',
    description='Web graphical user interface for NWB conversion and exploring',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Luiz Tauffer and Vinicius Camozzato Vaz',
    author_email='luiz@taufferconsulting.com',
    url='https://github.com/catalystneuro/nwb-web-gui',
    keywords='nwb',
    packages=find_packages(),
    package_data={
        'assets': ['*'],
        'static': ['*']
    },
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        'console_scripts': ['nwbgui=nwb_web_gui.cmd_line:cmd_line_shortcut']
    }
)



from flask import render_template, request
from flask import current_app as app


@app.route('/')
def home():
    """Landing page."""
    return render_template('home.html')

def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()

@app.route('/shutdown', methods=['POST', 'GET'])
def shutdown():
    shutdown_server()
    return 'Server down...'



import configparser
import os


class Config:
    """Base config."""
    # SECRET_KEY = 'SECRET_KEY'
    SESSION_COOKIE_NAME = 'SESSION_COOKIE_NAME'
    STATIC_FOLDER = 'static'
    TEMPLATES_FOLDER = 'templates'
    SEND_FILE_MAX_AGE_DEFAULT = 0

    NWB_GUI_CONVERTER_CLASS = 'example'  # default example converter


class ConfigProduction(Config):
    FLASK_ENV = 'production'
    DEBUG = False
    TESTING = False
    DATABASE_URI = 'DATABASE_URI_PRODUCTION'


class ConfigDev(Config):
    FLASK_ENV = 'development'
    DEBUG = True
    TESTING = True
    DATABASE_URI = 'DEV_DATABASE_URI'

    # The following variables are recovered by the app from ENV variables
    # In Development, we get them from a .ini file and set the ENV vars
    # In production, these ENV vars should be set in other ways
    parser = configparser.ConfigParser()
    parser.read('config.ini')

    if 'PATH' in parser.sections():
        os.environ['NWB_GUI_ROOT_PATH'] = parser['PATH']['NWB_GUI_ROOT_PATH']

    if 'CONVERTER' in parser.sections():
        os.environ['NWB_GUI_CONVERTER_MODULE'] = parser['CONVERTER']['NWB_GUI_CONVERTER_MODULE']
        os.environ['NWB_GUI_CONVERTER_CLASS'] = parser['CONVERTER']['NWB_GUI_CONVERTER_CLASS']

    if 'DASHBOARD' in parser.sections():
        os.environ['NWB_GUI_DASHBOARD_MODULE'] = parser['DASHBOARD']['NWB_GUI_DASHBOARD_MODULE']
        os.environ['NWB_GUI_DASHBOARD_CLASS'] = parser['DASHBOARD']['NWB_GUI_DASHBOARD_CLASS']

    if 'SECRETS' in parser.sections():
        os.environ['NWB_GUI_SECRET_KEY'] = parser['SECRETS']['NWB_GUI_SECRET_KEY']


config = {
    'dev': ConfigDev,
    'prod': ConfigProduction,
    'default': ConfigDev,
}



import dash
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State, ALL
import dash_bootstrap_components as dbc
import numpy as np
import json
import yaml
import base64
from json_schema_to_dash_forms.forms import SchemaFormContainer
from pathlib import Path
import flask
from io import StringIO
from contextlib import redirect_stdout
import threading
import time


class ConverterForms(html.Div):
    def __init__(self, parent_app, converter_class):
        """
        Forms to interface user input with NWB converters.

        INPUT:
        ------
        parent_app : running Dash app
        converter : NWB converter class
        """
        super().__init__([])
        self.parent_app = parent_app
        self.converter_class = converter_class
        self.export_controller = False
        self.convert_controller = False
        self.get_metadata_controller = False
        self.conversion_messages = ''
        self.conversion_msg_controller = True
        self.msg_buffer = StringIO()

        self.downloads_path = Path(__file__).parent.parent.parent.parent.parent.absolute() / 'downloads'
        self.root_path = Path(self.parent_app.server.config['NWB_GUI_ROOT_PATH'])

        if not self.downloads_path.is_dir():
            self.downloads_path.mkdir()

        self.source_json_schema = converter_class.get_source_schema()
        self.conversion_json_schema = converter_class.get_conversion_options_schema()

        # Source data Form
        self.source_forms = SchemaFormContainer(
            id='sourcedata',
            schema=self.source_json_schema,
            parent_app=self.parent_app,
            root_path=self.root_path
        )
        
        # Conversion Option Form
        self.conversion_options_forms = SchemaFormContainer(
            id='conversiondata',
            schema=self.conversion_json_schema,
            parent_app=self.parent_app,
            root_path=self.root_path
        )

        self.metadata_forms = SchemaFormContainer(
            id='metadata',
            schema=dict(),
            parent_app=self.parent_app,
            root_path=self.root_path
        )
        self.style = {'background-color': '#f0f0f0', 'min-height': '100vh'}

        self.children = [
            dbc.Container([
                dbc.Row([
                    html.Br(),
                    dbc.Col(self.source_forms, width={'size': 12}),
                    dbc.Col(
                        dbc.Button('Get Metadata Form', id='get_metadata_btn', color='dark'),
                        style={'justify-content': 'left', 'text-align': 'left', 'margin-top': '1%'},
                        width={'size': 4}
                    )
                ]),
                dbc.Row([
                    dbc.Col(
                        dcc.Upload(dbc.Button('Load Metadata', color='dark'), id='button_load_metadata', style={'display': 'none'}),
                        width={'size': 2},
                        style={'justify-content': 'left', 'text-align': 'left', 'margin-top': '1%'},
                    ),
                    dbc.Col(
                        html.Div([
                            dbc.Button('Export Metadata', id='button_export_metadata', color='dark', style={'display': 'none'}),
                            dbc.Popover(
                                [
                                    dbc.PopoverBody([
                                        html.Div([
                                            html.A(
                                                dbc.Button("Download as JSON", id='button_export_json', color="link"),
                                                href='/downloads/exported_metadata.json'
                                            ),
                                            html.A(
                                                dbc.Button("Download as YAML", id='button_export_yaml', color="link"),
                                                href='/downloads/exported_metadata.yaml'
                                            )
                                        ])
                                    ])
                                ],
                                id="popover_export_metadata",
                                target='button_export_metadata',
                                is_open=False,
                                placement='top',
                            )
                        ]),
                        width={'size': 2},
                        style={'justify-content': 'left', 'text-align': 'left', 'margin-top': '1%'},
                    ),
                    dbc.Col(
                        dbc.Button('Refresh', id='button_refresh', color='dark', style={'display': 'none'}),
                        width={'size': 2},
                        style={'justify-content': 'left', 'text-align': 'left', 'margin-top': '1%'},
                    )
                ]),
                dbc.Row([
                    dbc.Col(
                        dbc.Alert(
                            children=[],
                            id="alert_required_source",
                            dismissable=True,
                            is_open=False,
                            color='danger'
                        )
                    ),
                    dbc.Col(
                        dbc.Alert(
                            children=[],
                            id="alert_required",
                            dismissable=True,
                            is_open=False,
                            color='danger'
                        )
                    )
                ]),
                dbc.Row(
                    dbc.Col(id='metadata-col', width={'size': 12}),
                    style={'margin-top': '1%', 'margin-bottom': '10px'}
                ),
                dbc.Row(
                    dbc.Col(self.conversion_options_forms, id='conversion-col', width={'size': 12}, style={'display': 'none'}),
                ),
                dbc.Row(
                    dbc.Col(
                        dbc.InputGroup(
                            [
                                dbc.InputGroupAddon("Output file: ", addon_type="prepend"),
                                dbc.Input(id="output-nwbfile-name", placeholder="filename.nwb"),
                                dbc.InputGroupAddon(
                                    dbc.Button('Run Conversion', id='button_run_conversion'),
                                    addon_type="append",
                                ),
                            ]
                        ),
                        width={'size': 11}
                    ),
                    style={'text-align': 'left', 'margin-top': '1%', 'display': 'none'},
                    id='row_output_conversion'
                ),
                dbc.Row([
                    dbc.Col(
                        dbc.Alert(
                            children=[],
                            id="alert-required-conversion",
                            dismissable=True,
                            is_open=False,
                            color='danger'
                        )
                    )
                ]),
                dbc.Textarea(
                    id='text-conversion-results',
                    className='string_input',
                    bs_size="lg",
                    readOnly=True,
                    style={'font-size': '16px', 'display': 'none'}
                ),
                dbc.Button(id='pause_loop', style={'display': 'none'}),
                dcc.Interval(id='interval-text-results', max_intervals=0, interval=500),
                html.Br(),
                html.Div(id='export-output', style={'display': 'none'}),
                html.Div(id='export-input', style={'display': 'none'}),
                dbc.Button(id='get_metadata_done', style={'display': 'none'}),
            ], style={'min-height': '110vh'})
        ]

        @self.parent_app.callback(
            [
                Output("popover_export_metadata", "is_open"),
                Output('alert_required', 'is_open'),
                Output('alert_required', 'children'),
            ],
            [Input('metadata-output-update-finished-verification', 'children')],
            [
                State("popover_export_metadata", "is_open"),
                State('alert_required', 'is_open')
            ]
        )
        def export_metadata(trigger, fileoption_is_open, req_is_open):
            """
            Export Metadata Form data to JSON and YAML file
            This function is triggered when metadata internal dict is updated
            and export controller is setted to true.
            If export controller is not setted to true but the metadata internal dict was updated
            the function will return the current application state
            """
            # Prevent default
            if not self.export_controller or not trigger:
                return fileoption_is_open, req_is_open, []

            if self.export_controller and fileoption_is_open:
                self.export_controller = False
                return not fileoption_is_open, req_is_open, []

            alerts, output = self.metadata_forms.data_to_nested()

            # If required fields missing return alert
            if alerts is not None:
                return fileoption_is_open, not req_is_open, alerts

            # Make temporary files on server side
            # JSON
            exported_file_path = self.downloads_path / 'exported_metadata.json'
            with open(exported_file_path, 'w') as outfile:
                json.dump(output, outfile, indent=4)

            # YAML
            exported_file_path = self.downloads_path / 'exported_metadata.yaml'
            with open(exported_file_path, 'w') as outfile:
                yaml.dump(output, outfile, default_flow_style=False)

            return not fileoption_is_open, req_is_open, []

        @self.parent_app.callback(
            [
                Output('metadata-external-trigger-update-internal-dict', 'children'),
                Output('conversiondata-external-trigger-update-internal-dict', 'children')
            ],
            [
                Input('button_export_metadata', 'n_clicks'),
                Input('button_run_conversion', 'n_clicks')
            ],
        )
        def update_internal_metadata(click_export, click_conversion):
            """
            Trigger metadata internal dict update and then:
            1) set export_controller to true, when exporting to json/yaml
            2) set convert_controller to true, when running conversion
            """
            ctx = dash.callback_context
            if not ctx.triggered:
                return dash.no_update
            else:
                button_id = ctx.triggered[0]['prop_id'].split('.')[0]
                if button_id == 'button_export_metadata':
                    self.export_controller = True
                    self.convert_controller = False
                elif button_id == 'button_run_conversion':
                    self.export_controller = False
                    self.convert_controller = True
                return str(np.random.rand()), str(np.random.rand())

        @self.parent_app.callback(
            [
                Output('metadata-col', 'children'),
                Output('button_load_metadata', 'style'),
                Output('button_export_metadata', 'style'),
                Output('button_refresh', 'style'),
                Output('row_output_conversion', 'style'),
                Output('text-conversion-results', 'style'),
                Output('get_metadata_done', 'n_clicks'),
                Output('alert_required_source', 'is_open'),
                Output('alert_required_source', 'children'),
                Output('conversion-col', 'style')
            ],
            [Input('sourcedata-output-update-finished-verification', 'children')],
            [
                State('alert_required_source', 'is_open'),
                State('button_load_metadata', 'style'),
                State('button_export_metadata', 'style'),
                State('button_refresh', 'style'),
                State('row_output_conversion', 'style'),
                State('text-conversion-results', 'style'),
                State('conversion-col', 'style')
            ]
        )
        def get_metadata(trigger, alert_is_open, *styles):
            """
            Render Metadata forms based on Source Data Form
            This function is triggered when sourcedata internal dict is updated
            and get metadata controller is setted to true.
            If get metadata controller is not setted to true but the sourcedata
            internal dict was updated the function will return the current
            application state
            """
            if not trigger or not self.get_metadata_controller:
                # If metadata forms defined reset to default state
                if self.metadata_forms.children_forms:
                    self.metadata_forms.children_forms = []
                    self.metadata_forms.children = self.metadata_forms.children_triggers
                    self.metadata_forms.data = dict()
                    self.metadata_forms.schema = dict()
                return [self.metadata_forms, styles[0], styles[1], styles[2], styles[3], styles[4], None, alert_is_open, [], styles[5]]

            # Get forms data
            alerts, source_data = self.source_forms.data_to_nested()

            if alerts is not None:
                return [self.metadata_forms, styles[0], styles[1], styles[2], styles[3], styles[4], None, True, alerts, styles[5]]

            self.get_metadata_controller = False

            # Get metadata schema from converter
            self.converter = self.converter_class(source_data=source_data)
            self.metadata_json_schema = self.converter.get_metadata_schema()

            # Get metadata data from converter
            self.metadata_json_data = self.converter.get_metadata()

            if self.metadata_forms.children_forms:
                # Clean form children if exists to render new one
                self.metadata_forms.children_forms = []

            self.metadata_forms.schema = self.metadata_json_schema
            self.metadata_forms.construct_children_forms()
            self.metadata_forms.update_data(data=self.metadata_json_data)

            return [
                self.metadata_forms, {'display': 'block'}, {'display': 'block'},
                {'display': 'block'}, {'display': 'block'}, {'font-size': '16px', 'display': 'block', 'height': '100%', "min-height": "200px", "max-height": "600px"},
                1, alert_is_open, [], {'display': 'block'}
            ]

        @self.parent_app.callback(
            Output('sourcedata-external-trigger-update-internal-dict', 'children'),
            [Input('get_metadata_btn', 'n_clicks')]
        )
        def update_internal_sourcedata(click):
            """Update sourcedata internal dictionary to Get Metadata Forms from it"""
            if click:
                self.get_metadata_controller = True
                return str(np.random.rand())

        @self.parent_app.callback(
            Output({'type': 'external-trigger-update-links-values', 'index': 'metadata-external-trigger-update-links-values'}, 'children'),
            [Input('button_refresh', 'n_clicks')]
        )
        def refresh_forms_links(click):
            if click:
                return str(np.random.rand())

        @self.parent_app.callback(
            Output({'type': 'external-trigger-update-forms-values', 'index': 'metadata-external-trigger-update-forms-values'}, 'children'),
            [
                Input('button_load_metadata', 'contents'),
                Input('get_metadata_done', 'n_clicks')
            ],
            [State('button_load_metadata', 'filename')]
        )
        def update_forms_values_metadata(contents, click, filename):
            """
            Updates forms values (except links) when:
            - Forms are created (receives metadata dict from Converter)
            - User uploads metadata json / yaml file
            """
            ctx = dash.callback_context
            trigger_source = ctx.triggered[0]['prop_id'].split('.')[0]

            if trigger_source != 'button_load_metadata' and click is None:
                output = []
                return output

            if trigger_source != 'button_load_metadata' and click is not None:
                output = str(np.random.rand())
                return output

            content_type, content_string = contents.split(',')
            filename_extension = filename.split('.')[-1]

            # Update SchemaFormContainer internal data dictionary
            if filename_extension == 'json':
                bs4decode = base64.b64decode(content_string)
                json_string = bs4decode.decode('utf8').replace("'", '"')
                self.metadata_json_data = json.loads(json_string)
                self.metadata_forms.update_data(data=self.metadata_json_data)
            elif filename_extension in ['yaml', 'yml']:
                bs4decode = base64.b64decode(content_string)
                yaml_data = yaml.load(bs4decode, Loader=yaml.BaseLoader)
                self.metadata_json_data = yaml_data
                self.metadata_forms.update_data(data=self.metadata_json_data)
            # Trigger update of React components
            output = str(np.random.rand())

            return output

        @self.parent_app.server.route('/downloads/<path:filename>')
        def download_file(filename):
            return flask.send_from_directory(
                directory=self.downloads_path,
                filename=filename,
                as_attachment=True
            )

        @self.parent_app.callback(
            [
                Output('interval-text-results', 'max_intervals'),
                Output('alert-required-conversion', 'is_open'),
                Output('alert-required-conversion', 'children'),
            ],
            [
                Input('metadata-output-update-finished-verification', 'children'),
                Input('pause_loop', 'n_clicks')
            ],
            [
                State('output-nwbfile-name', 'value'),
                State('alert-required-conversion', 'is_open')
            ]
        )
        def trigger_conversion(trigger, pause, output_nwbfile, alert_is_open):
            ctx = dash.callback_context
            trigger_source = ctx.triggered[0]['prop_id'].split('.')[0]

            if trigger_source == 'metadata-output-update-finished-verification' and self.convert_controller:
                # run conversion
                alerts, metadata = self.metadata_forms.data_to_nested()

                _, conversion_options_data = self.conversion_options_forms.data_to_nested() # use this metadata to conversion
                
                if alerts is not None:
                    return 0, True, alerts

                nwbfile_path = output_nwbfile

                self.msg_buffer.truncate(0)

                self.t = threading.Thread(target=self.conversion, daemon=True, args=(metadata, nwbfile_path, conversion_options_data))
                self.t.start()

                self.conversion_msg_controller = True
                return -1, False, [] # run loop

            elif trigger_source == 'pause_loop' and pause is not None:
                # Pause interval component that reads conversion messages and terminate conversion thread
                if self.t.is_alive():
                    self.t.terminate()
                return 0, False, []

            return dash.no_update

        @self.parent_app.callback(
            [
                Output('text-conversion-results', 'value'),
                Output('pause_loop', 'n_clicks')
            ],
            [
                Input('interval-text-results', 'n_intervals')
            ]
        )
        def update_conversion_messages(n_intervals):
            self.conversion_messages = self.msg_buffer.getvalue()

            if self.conversion_msg_controller:
                return self.conversion_messages, None
            return self.conversion_messages, 1

    def conversion(self, metadata, nwbfile_path, conversion_options):
        try:
            with redirect_stdout(self.msg_buffer):
                self.converter.run_conversion(
                    metadata=metadata,
                    nwbfile_path=nwbfile_path,
                    save_to_file=True,
                    conversion_options=conversion_options
                )
        except Exception as e:
            self.msg_buffer.write(str(e))
        finally:
            time.sleep(0.1)
            self.convert_controller = False
            self.conversion_msg_controller = False
            



from json_schema_to_dash_forms.utils import make_filebrowser_modal, FileBrowserComponent
from flask import current_app as app
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, ALL, MATCH
import dash
from pathlib import Path
from pynwb import NWBHDF5IO
import subprocess
import multiprocessing
from nwbwidgets import nwb2widget
import nbformat as nbf
import os
import time


def call_voila(run_cmd):
    """Call voila processes"""
    subp = subprocess.run(run_cmd, shell=True, capture_output=True)
    print(subp.stdout.decode())


class Viewer(html.Div):
    def __init__(self, parent_app):
        super().__init__([])

        self.parent_app = parent_app
        self.processes_list = []
        self.nwb_file = None

        # Viewer page layout
        self.filebrowser = FileBrowserComponent(parent_app=parent_app, id_suffix='viewer')

        self.children = [
            dbc.Toast(
                id="toast_loadedfile_viewer",
                is_open=False,
                dismissable=False,
                duration=5000
            ),
            html.Br(),
            self.filebrowser,
            html.Br(),
            html.Div(id='voila_div', style={'justify-content': 'center', 'text-align': 'center'}),
        ]

        @self.parent_app.callback(
            [
                Output("toast_loadedfile_viewer", "is_open"),
                Output("toast_loadedfile_viewer", "style"),
                Output("toast_loadedfile_viewer", "children"),
                Output('voila_div', 'children'),
                Output("button_file_browser_viewer", 'n_clicks')
            ],
            [Input('submit-filebrowser-viewer', component_property='n_clicks')],
            [State('chosen-filebrowser-viewer', 'value')]
        )
        def submit_nwb(click, input_value):

            if not click:
                return dash.no_update

            if not any(d['key'] == input_value for d in self.filebrowser.paths_tree):
                # If not selected from explorer
                nwb_path = Path(input_value)
            else:
                # If selected from explorer get absolute path from config
                nwb_path = Path(app.config['NWB_GUI_ROOT_PATH']).parent / Path(input_value)

            style = {
                "position": "fixed", "top": 180, "left": 10, "width": 350,
                "background-color": "#955", "color": "#ffffff", "solid": True
            }
            ctx = dash.callback_context
            source = ctx.triggered[0]['prop_id'].split('.')[0]

            if source == 'submit-filebrowser-viewer':
                if nwb_path.is_file() and str(nwb_path).endswith('.nwb'):
                    self.nwb_path = nwb_path
                    voila_address = self.run_explorer()
                    time.sleep(5)
                    iframe = dbc.Container(html.Iframe(style={"min-width": "100%", 'max-width': '100%', "min-height": "100vh"}, src=voila_address))
                    style.update({"background-color": "#287836"})
                    return True, style, 'NWB file loaded, widgets started', iframe, 1
                else:
                    style.update({"background-color": "#955"})
                    return True, style, 'Must be a NWB file', '', None
            return False, '', '', '', None

    def kill_processes(self):
        """"""
        if len(self.processes_list) > 0:
            for p in self.processes_list:
                p.terminate()
                p.join()
                p.close()
                print('Terminated: ', p)
            self.processes_list = []

    def run_explorer(self):
        """"""
        self.kill_processes()
        self.aux_notebook = self.create_notebook()

        settings_voila = {'allow_origin': '*', 'headers': {
            'Content-Security-Policy': 'frame-ancestors http://localhost:5000'}}

        run_cmd = ' '.join([
            "voila",
            """ "{}" """.format(str(self.aux_notebook)),
            "--no-browser",
            "--debug",
            "--enable_nbextensions=True",
            "--port=8866",
            "--strip_sources=" + str(True),
            """--Voila.tornado_settings="{}" """.format(settings_voila)
        ])

        proc_voila = multiprocessing.Process(target=call_voila, args=(run_cmd,))
        proc_voila.start()
        self.processes_list.append(proc_voila)
        voila_address = 'http://localhost:8866'

        return voila_address

    def create_notebook(self):
        '''Create aux notebook file for voila render'''

        code = """
            from nwbwidgets import nwb2widget
            import pynwb
            from pathlib import Path
            import os
            fpath = os.path.join(r'""" + str(self.nwb_path) + """')
            io = pynwb.NWBHDF5IO(fpath, 'r', load_namespaces=True)
            nwb = io.read()
            nwb2widget(nwb)
        """

        nb = nbf.v4.new_notebook()
        nb['cells'] = [nbf.v4.new_code_cell(code)]
        fname = 'temp_voila_notebook.ipynb'
        path_aux_notebook = Path.cwd() / fname
        nbf.write(nb, path_aux_notebook)

        return path_aux_notebook
